import torch
import torch.nn as nn
import torch.nn.init as init

def kaiming_init(m):
	if isinstance(m, (nn.Linear, nn.Conv2d)):
		init.kaiming_normal_(m.weight)
		if m.bias is not None:
			m.bias.data.fill_(0)
	elif isinstance(m, (nn.BatchNorm1d, nn.BatchNorm2d)):
		m.weight.data.fill_(1)
		if m.bias is not None:
			m.bias.data.fill_(0)

class View(nn.Module):
	def __init__(self, size):
		super(View, self).__init__()
		self.size = size

	def forward(self, tensor):
		return tensor.view(self.size)

class Encoder(nn.Module):
  def __init__(self, z_dim=6):
    super(Encoder, self).__init__()
    self.z_dim = z_dim
    self.encoder = nn.Sequential(
			nn.Conv2d(1, 32, 4, 2, 1),
			nn.ReLU(True),
			nn.Conv2d(32, 32, 4, 2, 1),
			nn.ReLU(True),
			nn.Conv2d(32, 64, 4, 2, 1),
			nn.ReLU(True),
			nn.Conv2d(64, 64, 4, 2, 1),
			nn.ReLU(True),
			nn.Conv2d(64, 256, 4, 1),
			nn.ReLU(True),
			View((-1, 256*1*1)),
			nn.Linear(256, z_dim * 2)
		)

    # init
    for block in self._modules:
      for m in self._modules[block]:
        kaiming_init(m)

  def forward(self, _x):
    z = self.encoder(_x)

    return z

class Decoder(nn.Module):
  def __init__(self, z_dim=6):
    super(Decoder, self).__init__()
    self.z_dim = z_dim
    self.decoder = nn.Sequential(
			nn.Linear(self.z_dim, 256),
			View((-1, 256, 1, 1)),
			nn.ReLU(True),
			nn.ConvTranspose2d(256, 64, 4),
			nn.ReLU(True),
			nn.ConvTranspose2d(64, 64, 4, 2, 1),
			nn.ReLU(True),
			nn.ConvTranspose2d(64, 32, 4, 2, 1),
			nn.ReLU(True),
			nn.ConvTranspose2d(32, 32, 4, 2, 1),
			nn.ReLU(True),
			nn.ConvTranspose2d(32, 1, 4, 2, 1)
		)

    # init
    for block in self._modules:
      for m in self._modules[block]:
        kaiming_init(m)

  def forward(self, _z):
    x_recon = self.decoder(_z)

    return x_recon

class Dataset:
  def __init__(self, _data):
    self.data = _data['data']
    self.values = _data['values']
    self.labels = _data['labels']

  def __len__(self):
    return len(self.data)

  def __getitem__(self, _i):
    return (self.data[_i], self.values[_i], self.labels[_i])