"""
Multitype disentanglement through categorical labelling and semi-supervised continuous finetuning
"""

import numpy as np
import torch
from torch.utils.data import DataLoader
from tqdm import tqdm
import board

from mtd_VAE_14 import dSprites_Model
from dsprites_AE import Encoder, Decoder, Dataset

DEFAULT_DIR = './models'
DEFAULT_EXT = '.md'

def correlation(_labels, _mus):
  return np.abs(np.corrcoef(_labels, _mus.detach().squeeze().cpu())[0][1])

def main(_rank, _nprocs, _args):
  device = 'cuda:' + str(_rank)
  log = _args['log']

  train_dataloader = DataLoader(_args["train_dataset"], batch_size=log['S_BATCH'], num_workers=6, persistent_workers=True, pin_memory=True, shuffle=True)

  model = _args['model'].to(device)
  optimizer = torch.optim.Adam(model.parameters(), lr=log['LR'] * _nprocs)
  scheduler = torch.optim.lr_scheduler.StepLR(optimizer, log['S_SZ'], log['S_MUL'])

  style1_labels_mean = np.array([_args['labels_info'].item()[key][2] for key in _args['labels_used'][0]])
  style2_labels_mean = np.array([_args['labels_info'].item()[key][2] for key in _args['labels_used'][1]])

  label_offset = min(_args['labels_used'][0])

  nm_iters = int(len(train_dataloader.dataset) / (train_dataloader.batch_size * _nprocs))
  model.train()
  for epoch in range(log['EPOCHS'] + 2):    
    if _rank == 2:
      pbar = tqdm(total=nm_iters)
    train_losses = []
    train_emb_losses = []
    train_cl_losses = []
    train_sc_losses = []
    train_probnll_losses = []

    progress = min(1.0, max(0.0, (epoch - 5) / (log['EPOCHS'] - 5)))
    
    for batch in train_dataloader:
      optimizer.zero_grad()

      x = batch[0].to(device).float()
      v = None
      l = batch[2].to(device) - label_offset

      loss_mean, emb_loss, cl_loss, sc_loss, probnll_loss = model(x, v, l, progress)
      loss_mean.backward()
      
      optimizer.step()

      if _rank == 2:
        train_losses.append(loss_mean.item())
        train_emb_losses.append(emb_loss.item())
        train_cl_losses.append(cl_loss.item())
        train_sc_losses.append(sc_loss.item())
        train_probnll_losses.append(probnll_loss.item())

        pbar.update(1)

    scheduler.step()
    
    if _rank == 2:
      pbar.close()

      log('norm/xaxis', torch.cat((model.style1_mu, torch.exp(model.style1_logvar*0.5)), dim=-1))
      log('norm/yaxis', torch.cat((model.style2_mu, torch.exp(model.style2_logvar*0.5)), dim=-1))
      log.step()
      
      print(f'{epoch} \
        - R Loss: {np.mean(train_losses)} \
        - E Loss: {np.mean(train_emb_losses)} \
        - C Loss: {np.mean(train_cl_losses)} \
        - SC Loss: {np.mean(train_sc_losses)} \
        - PN Loss: {np.mean(train_probnll_losses)} \
        - corr1: {correlation(style1_labels_mean, model.style1_mu)} \
        - corr2: {correlation(style2_labels_mean, model.style2_mu)} \
      \n')

if __name__ == "__main__":
  log = board.RunLog('r', _log='http://0.0.0.0:19364/api/runinfo/')

  log['EPOCHS'] = 24
  log['LR'] = 1e-3
  log['ZDIM'] = 20
  log['S_MUL'] = 0.1
  log['S_SZ'] = 100
  log['S_BATCH'] = 512
  log['STYLE_LEN'] = 1
  log['CONTENT_LEN'] = log['ZDIM'] - log['STYLE_LEN'] * 2

  #### 14 ####
  log['M_LOSS_RECON'] = (600.0, 1200.0)
  log['M_LOSS_BETA'] = (0.01, 0.1)
  log['M_LOSS_SC'] = (600.0, 800.0)
  log['M_LOSS_NLL'] = (600.0, 1000.0)
  log['M_LOSS_KL'] = (0.5,  5.0)
  log['M_LOSS_EMB'] = (5.0, 0.1)
  log['M_LOSS_CL'] = (0.1, 0.001)
  log['M_LOSS_SD'] = (0.001, 0.001)

  data = np.load('data/dsprites_ap_full.npz', encoding='bytes', mmap_mode='r', allow_pickle=True)
  log['STYLE1_LABELNUM'] = len(data['labels_used'][0])
  log['STYLE2_LABELNUM'] = len(data['labels_used'][1])

  encoder = Encoder(z_dim = log['ZDIM'])
  decoder = Decoder(z_dim = log['ZDIM'])
  model = dSprites_Model(log, encoder, decoder)
  
  train_dataset = Dataset({'data': data['imgs'], 'values': data['latents_values'], 'labels': data['labels']})

  try:
    args = {
      'log': log,
      'model': model,
      'train_dataset': train_dataset,
      'labels_info': data['labels_info'],
      'labels_used': data['labels_used']
    }
    main(2, 1, args)
  except KeyboardInterrupt:
    pass
  
  log.save()
  s = input('Do you want to save the model? ')
  if s == 'y':
    torch.save(model.state_dict(), './models/' + log.fullname + '.md')
